/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *  MA 02110-1301, USA.
 *
 *  Copyright © 2006 Pierre Habouzit
 */
/*
 * This file is part of mutt-ng, see http://www.muttng.org/.
 * It's licensed under the GNU General Public License,
 * please see the file GPL in the top level source directory.
 */

#include "lib-lib.h"

rx_t *rx_compile(const char *s, int flags)
{
    rx_t *pp = p_new(rx_t, 1);

    pp->pattern = m_strdup(s);
    pp->rx = p_new(regex_t, 1);

    if (REGCOMP(pp->rx, NONULL(s), flags) != 0) {
        rx_delete(&pp);
    }

    return pp;
}

int rx_validate(const char *s, char *errbuf, ssize_t errlen)
{
    regex_t re;
    int res;

    p_clear(&re, 1);
    res = REGCOMP(&re, NONULL(s), 0);
    if (res) {
        regerror(res, &re, errbuf, errlen);
    }
    regfree(&re);

    return res;
}

void rx_set_template(rx_t *rx, const char *tpl)
{
    const char *p = tpl;

    m_strreplace(&rx->tpl, tpl);
    rx->nmatch = 0;

    while ((p = strchr(p, '%'))) {
        if (isdigit(*++p)) {
            int n = strtol(p, (char **)&p, 10);
            rx->nmatch = MAX(n, rx->nmatch);
        } else {
            if (*p == '%')
                p++;
        }
    }

    rx->nmatch++;     /* match 0 is always the whole expr */
}

void rx_wipe(rx_t *rx)
{
    p_delete(&rx->pattern);
    regfree(rx->rx);
    p_delete(&rx->rx);
    p_delete(&rx->tpl);
}

int rx_list_match(rx_t *l, const char *s)
{
    if (m_strisempty(s))
        return 0;

    while (l) {
        if (!REGEXEC(l->rx, s))
            return 1;
        l = l->next;
    }

    return 0;
}

int rx_list_match2(rx_t *l, const char *s, char *dst, int dlen)
{
    static regmatch_t *pmatch = NULL;
    static int nmatch = 0;
    int pos = 0;

    if (m_strisempty(s))
        return 0;

    for (; l; l = l->next) {
        if (l->nmatch > nmatch) {
            p_realloc(&pmatch, l->nmatch);
            nmatch = l->nmatch;
        }

        if (regexec(l->rx, s, l->nmatch, pmatch, 0) == 0) {
            /* Copy template into dst, with substitutions. */
            const char *p = l->tpl, *q;

            for (q = strchr(p, '%'); q; q = strchr(p + 1, '%')) {
                int n;

                pos += m_strncpy(dst + pos, dlen - pos, p, q - p);

                if (!isdigit((unsigned char)q[1])) {
                    p = q + (q[1] == '%');
                    continue;
                }

                n = strtol(q + 1, (char **)&p, 10); /* find pmatch index */
                pos += m_strncpy(dst + pos, dlen - pos, s + pmatch[n].rm_so,
                                 pmatch[n].rm_eo - pmatch[n].rm_so);
            }

            pos += m_strcpy(dst + pos, dlen - pos, p);
            return 1;
        }
    }

    return 0;
}

rx_t **rx_lookup(rx_t **l, const char *pat)
{
    if (m_strisempty(pat))
        return NULL;

    while (*l) {
        if (!m_strcmp((*l)->pattern, pat))
            return l;
        l = &(*l)->next;
    }

    return l;
}

void rx_list_add(rx_t **l, rx_t *rxp)
{
    l = rx_lookup(l, rxp->pattern);
    if (*l) {
        rx_t *r = rx_list_pop(l);
        rx_delete(&r);
    }
    rx_list_push(l, rxp);
}

void rx_list_add2(rx_t **l, rx_t **rxp)
{
    l = rx_lookup(l, (*rxp)->pattern);
    if (*l) {
        rx_t *r = rx_list_pop(l);
        rx_delete(&r);
    }
    if (m_strisempty((*rxp)->tpl)) {
        rx_delete(rxp);
    } else {
        rx_list_push(l, *rxp);
    }
}

void rx_list_remove(rx_t **l, const rx_t *r)
{
    l = rx_lookup(l, r->pattern);
    if (*l) {
        rx_t *tmp = rx_list_pop(l);
        rx_delete(&tmp);
    }
}

int rx_sanitize_string(char *dst, ssize_t n, const char *src)
{
    while (*src) {
        if (n <= 1)
            break;

        /* these characters must be escaped in regular expressions */
        if (strchr("^.[$()|*+?{\\", *src)) {
            if (n <= 2)
                break;

            *dst++ = '\\';
            n--;
        }

        *dst++ = *src++;
        n--;
    }

    *dst = '\0';

    return *src ? -1 : 0;
}
